#include "stdafx.h"
#include "ObjFile.h"


static const std::string  EMPTY_STRING;

#define IS_VALID_PTR(P)   (IsValidPtr((P), sizeof(*(P))))


//=============================================================================


//! t@C std::vector<BYTE> ɓǂݍށB
static bool LoadFile(const wchar_t* i_obj_path, std::vector<BYTE>& o_data)
{
  bool            retval      = false;
  HANDLE          file_handle = INVALID_HANDLE_VALUE;
  LARGE_INTEGER   file_size   = { 0 };
  DWORD           read_size   = 0;

  do
  {
    file_handle = CreateFileW(
                    i_obj_path,
                    GENERIC_READ,
                    FILE_SHARE_READ,
                    NULL,
                    OPEN_EXISTING,
                    FILE_ATTRIBUTE_NORMAL,
                    NULL
                  );
    if( file_handle == INVALID_HANDLE_VALUE )
    {
      _RPTF2(_CRT_WARN,
             "unable to open file[%s] : 0x%08X.\n",
             i_obj_path, GetLastError()
            );
      break;
    }
    
    if( GetFileSizeEx(file_handle, &file_size) == FALSE )
    {
      _RPTF1(_CRT_WARN, "unable to get file size : 0x%08X.\n", GetLastError());
      break;
    }
    
    if( file_size.HighPart
    ||  file_size.LowPart > LONG_MAX
    ||  file_size.LowPart == 0 )
    {
      _RPTF0(_CRT_WARN, "invalid file size\n");
      break;
    }
    
    o_data.resize(file_size.LowPart, 0);
    if( ReadFile(file_handle, &o_data[0], file_size.LowPart, &read_size, NULL)
        == FALSE )
    {
      _RPTF1(_CRT_WARN, "unable to read file : 0x%08X.\n", GetLastError());
      break;
    }
    
    retval = true;
  }
  while( 0 );

  if( file_handle != INVALID_HANDLE_VALUE )
  {
    CloseHandle(file_handle);
    file_handle = INVALID_HANDLE_VALUE;
  }

  return retval;
}


//=============================================================================


//! RXgN^Bobj t@CǂݍށB
ObjFile::ObjFile(const wchar_t* i_obj_path)
: m_is_opened(false),
  m_obj_data(),
  m_file_header_p(NULL),
  m_section_header_p(NULL),
  m_symbols_p(NULL),
  m_symbol_names()
{
  if( LoadFile(i_obj_path, m_obj_data) )
  {
    Analyze();
  }
}


//=============================================================================


//! RXgN^B obj f[^Rs[ĕێB
ObjFile::ObjFile(const void* i_obj_data_p, size_t i_size)
: m_is_opened(false),
  m_obj_data(),
  m_file_header_p(NULL),
  m_section_header_p(NULL),
  m_symbols_p(NULL),
  m_symbol_names()
{
  if( i_size && IsBadReadPtr(i_obj_data_p, i_size) == FALSE )
  {
    // i_obj_data_p Rs[B
    const BYTE* obj_data_p = reinterpret_cast<const BYTE*>(i_obj_data_p);
    m_obj_data.resize(i_size, 0);
    std::copy(obj_data_p, obj_data_p + i_size, m_obj_data.begin());
    
    Analyze();
  }
}


//=============================================================================


//! fXgN^
ObjFile::~ObjFile()
{
  // ɂ邱ƂȂ
}


//=============================================================================


//! JǂmFB
bool ObjFile::IsOpened() const
{
  return m_is_opened;
}


//=============================================================================


//! V{擾B
int ObjFile::GetSymbolCount() const
{
  int   retval  = 0;
  
  if( m_is_opened && m_file_header_p )
    retval = m_file_header_p->NumberOfSymbols;
  
  return retval;
}


//=============================================================================


//! V{̃|C^擾B
const IMAGE_SYMBOL* ObjFile::GetSymbolP(int i_index) const
{
  const IMAGE_SYMBOL* retval = NULL;
  
  do
  {
    if( !(0 <= i_index && i_index < GetSymbolCount()) )
    {
      _RPTF2(_CRT_WARN, "CfbNX(%d)͈(%d)OłB\n",
             i_index, GetSymbolCount()
            );
      break;
    }
    
    if( IS_VALID_PTR(m_symbols_p + i_index) == false )
    {
      _RPTF0(_CRT_WARN, "V{e[uĂ܂B\n");
      break;
    }
    
    retval = m_symbols_p + i_index;
  }
  while( 0 );
  
  return retval;
}


//=============================================================================


//! i_index Ԗ(0-based)̃V{擾B
const std::string& ObjFile::GetSymbolName(int i_index) const
{
  if( 0 <= i_index && i_index < GetSymbolCount() )
    return m_symbol_names[i_index];
  else
    return EMPTY_STRING;
}


//=============================================================================


//! i_index Ԗ(0-based)̔CV{擾B
const std::string& ObjFile::GetUndecoratedName(int i_index) const
{
  if( 0 <= i_index && i_index < GetSymbolCount() )
    return m_undecorated_names[i_index];
  else
    return EMPTY_STRING;
}


//=============================================================================


//! V{CĂ邩ǂׂB
bool ObjFile::IsDecorated(int i_index) const
{
  return GetSymbolName(i_index)[0] == '?';
}


//=============================================================================


//! V{O`V{̃CfbNX擾B
/**
 *  @return ꍇ͂̃V{̃CfbNXBȂꍇ
 *           -1 ԂB
 */
int ObjFile::FindSymbol(const char* i_symbol_name) const
{
  int   retval = -1;

  if( IsBadReadPtr(i_symbol_name, sizeof(char)) == FALSE )
  {
    for(int i = 0; i < GetSymbolCount(); i++)
    {
      if( IsExternalSymbol(i) && GetSymbolName(i) == i_symbol_name )
      {
        retval = i;
        break;
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! |C^ m_obj_data wĂ邩ǂׂB
bool ObjFile::IsValidPtr(const void* i_ptr, size_t i_size) const
{
  return m_obj_data.size()
      && &m_obj_data[0] <= i_ptr && i_ptr < &m_obj_data[0] + m_obj_data.size()
       ;
}


//=============================================================================


//! m_obj_data ̒g͂āAeoɒlZbgB
void ObjFile::Analyze()
{
  BYTE*       obj_data_p            = NULL;
  DWORD       long_name_table_size  = 0;
  const BYTE* long_name_table_p     = NULL;

  do
  {
    if( m_obj_data.size() == 0 )
    {
      _RPTF0(_CRT_WARN, "m_obj_data.size()  0 łB\n");
      break;
    }
    
    obj_data_p = &m_obj_data[0];

    // t@Cwb_
    m_file_header_p = reinterpret_cast<IMAGE_FILE_HEADER*>(obj_data_p);
    if( IS_VALID_PTR(m_file_header_p) == false )
    {
      _RPTF0(_CRT_WARN, "t@CĂ邩 COFF t@Cł͂܂B\n");
      break;
    }
    
    // Ó̊mFB
    if( m_file_header_p->Machine != IMAGE_FILE_MACHINE_I386
    &&  m_file_header_p->Machine != IMAGE_FILE_MACHINE_UNKNOWN )
    {
      _RPTF4(_CRT_WARN,
             "s COFF f[^łB\n"
             "  Ώۃ}V   : %08Xh\n"
             "  ZNV : %d\n"
             "  V{   : %d\n"
             "           : %08Xh\n",
             m_file_header_p->Machine,
             m_file_header_p->NumberOfSections,
             m_file_header_p->NumberOfSymbols,
             m_file_header_p->Characteristics
            );
      break;
    }
    
    // [ COFF łȂmFB
    if( m_file_header_p->Machine == IMAGE_FILE_MACHINE_UNKNOWN
    &&  m_file_header_p->NumberOfSections == IMPORT_OBJECT_HDR_SIG2 )
    {
      _RPTF0(_CRT_WARN, "C|[gp[ COFF t@CłB\n");
      break;
    }
    
    // ԍŏ̃ZNVwb_ւ̃|C^擾B
    m_section_header_p = reinterpret_cast<IMAGE_SECTION_HEADER*>(
                           m_file_header_p + 1
                         );
    if( IS_VALID_PTR(m_section_header_p) == false )
    {
      _RPTF0(_CRT_WARN, "t@CĂ邩 COFF t@Cł͂܂B\n");
      break;
    }
    
    // V{e[uւ̃|C^擾B
    m_symbols_p = reinterpret_cast<IMAGE_SYMBOL*>(
                    obj_data_p + m_file_header_p->PointerToSymbolTable
                  );
    if( IS_VALID_PTR(m_symbols_p) == false )
    {
      _RPTF0(_CRT_WARN, "t@CĂ邩 COFF t@Cł͂܂B\n");
      break;
    }
    
    // OV{e[uւ̃|C^擾B
    // V{e[ǔɑB
    long_name_table_p = reinterpret_cast<const BYTE*>(
                          m_symbols_p + m_file_header_p->NumberOfSymbols
                        );
    if( IsValidPtr(long_name_table_p, sizeof(DWORD)) == false )
    {
      _RPTF0(_CRT_WARN, "t@CĂ邩 COFF t@Cł͂܂B\n");
      break;
    }
    
    // OV{e[u̐擪ɂ̓e[uTCYB
    long_name_table_size = *reinterpret_cast<const DWORD*>(long_name_table_p);
    if( IsValidPtr(long_name_table_p, long_name_table_size) == false )
    {
      _RPTF0(_CRT_WARN, "t@CĂ邩 COFF t@Cł͂܂B\n");
      break;
    }
    
    // OV{e[ȕI̓t@C̏I[łȂƑʖځB
    if( long_name_table_p + long_name_table_size
        != obj_data_p + m_obj_data.size() )
    {
      _RPTF0(_CRT_WARN, "t@C̏I[sł\n");
      break;
    }
    
    // obt@̊mہB
    m_symbol_names.resize(m_file_header_p->NumberOfSymbols);
    m_undecorated_names.resize(m_file_header_p->NumberOfSymbols);
    
    // V{擾B
    for(size_t i = 0; i < m_file_header_p->NumberOfSymbols; i++)
    {
      const IMAGE_SYMBOL* symbol_p = m_symbols_p + i;
      if( IS_VALID_PTR(symbol_p) )
      {
        if( symbol_p->N.Name.Short )  // V{8ȓ
        {
          // '\0' ŏIĂȂꍇ(8)̏ꍇ
          if( symbol_p->N.ShortName[7] != '\0' )
          {
            char  name[sizeof(symbol_p->N.ShortName) + 1] = { 0 };
            memcpy(name, symbol_p->N.ShortName, sizeof(symbol_p->N.ShortName));
            m_symbol_names[i] = name;
          }
          // 8 '\0' ŏIĂꍇB
          else
          {
            m_symbol_names[i] = reinterpret_cast<const char*>(
                                  symbol_p->N.ShortName
                                );
          }
        }
        // OV{̏ꍇB
        else
        {
          const char* long_name = reinterpret_cast<const char*>(
                                    long_name_table_p + symbol_p->N.Name.Long
                                  );
          if( IsValidPtr(long_name, sizeof(symbol_p->N.ShortName) + 2) == false )
          {
            _RPTF0(_CRT_WARN, "V{e[uĂ܂B\n");
            break;
          }
          
          m_symbol_names[i] = long_name;
        }
        
        // C̎擾
        if( m_symbol_names[i][0] == '?' )
          UndecorateName(m_undecorated_names[i], m_symbol_names[i]);
        else
          m_undecorated_names[i] = m_symbol_names[i];
        
        i += symbol_p->NumberOfAuxSymbols;
      }
    }
    
    m_is_opened = true;
  }
  while( 0 );
}


//=============================================================================


// C^CCu̔J֐Bς݃V{̑B
extern "C" char* __unDNameEx(
  char*       o_undecorated_name,   // ς݂̖O̊i[
  const char* i_decorated_name,     // ς݃V{
  size_t      i_size,               // o_undecorated_name ̊i[\TCY
  void*       (*malloc_p)(size_t),  // ꎞ蓖Ă֐
  void        (*free_p)(void*),     // ꎞ֐
  void*       i_unknown,            // s
  unsigned    i_flags               // UnDecorateSymbolName()ƓtO
);


//! C擾B
bool ObjFile::UndecorateName(
  std::string&        o_undecorated_name,
  const std::string&  i_decorated_name
)
{
  bool              retval  = false;
  size_t            size    = 16;
  std::vector<char> undecorated_name;
  
  do
  {
    size *= 2;
    undecorated_name.resize(size, '\0');

    __unDNameEx(
      &undecorated_name[0],
      i_decorated_name.c_str(),
      size,
      malloc,
      free,
      NULL,
      0
    );
  }
  while( undecorated_name[size - 3] != '\0' );

  o_undecorated_name = &undecorated_name[0];
  
  return retval;
}


//=============================================================================


//! i_index Ԗ(0-based)̃ZNVwb_ւ̃|C^擾B
const IMAGE_SECTION_HEADER* ObjFile::GetSectionHeaderP(int i_index) const
{
  IMAGE_SECTION_HEADER*   retval = NULL;

  if( m_file_header_p
  &&  0 <= i_index && i_index < m_file_header_p->NumberOfSections
  &&  IS_VALID_PTR(m_section_header_p + i_index) )
  {
    retval = m_section_header_p + i_index;
  }
  
  return retval;
}


//=============================================================================


//! i_index Ԗ(0-based)̃ZNVf[^ւ̃|C^擾B
const BYTE* ObjFile::GetSectionDataP(int i_index) const
{
  const BYTE*                 retval            = NULL;
  const IMAGE_SECTION_HEADER* section_header_p  = GetSectionHeaderP(i_index);
  
  if( section_header_p && section_header_p->PointerToRawData
  &&  IsValidPtr(
        &m_obj_data[section_header_p->PointerToRawData],
        section_header_p->SizeOfRawData
      ) )
  {
    retval = &m_obj_data[section_header_p->PointerToRawData];
  }
  
  return retval;
}


//=============================================================================


//! i_index Ԗ(0-based)̃V{O`V{ǂׂB
bool ObjFile::IsExternalSymbol(int i_index) const
{
  bool                  retval    = false;
  const IMAGE_SYMBOL*   symbol_p  = GetSymbolP(i_index);
  
  if( symbol_p )
  {
    retval = symbol_p->StorageClass == IMAGE_SYM_CLASS_EXTERNAL
          && (symbol_p->SectionNumber > IMAGE_SYM_UNDEFINED || symbol_p->Value)
          ;
  }
  
  return retval;
}


//=============================================================================


//! i_index Ԗ(0-based)̃ZNV̍Ĕzu̐擾B
int ObjFile::GetRelocationCount(int i_index) const
{
  int                           retval            = 0;
  const IMAGE_SECTION_HEADER*   section_header_p  = GetSectionHeaderP(i_index);
  
  if( section_header_p )
    retval = section_header_p->NumberOfRelocations;
  
  return retval;
}


//=============================================================================


//! i_index Ԗ(0-based)̃ZNV̍Ĕzue[u擾B
const IMAGE_RELOCATION* ObjFile::GetRelocations(int i_index) const
{
  const IMAGE_RELOCATION*       retval            = NULL;
  const IMAGE_SECTION_HEADER*   section_header_p  = GetSectionHeaderP(i_index);
  
  if( section_header_p
  &&  section_header_p->NumberOfRelocations
  &&  IsValidPtr(
        &m_obj_data[section_header_p->PointerToRelocations],
        sizeof(IMAGE_RELOCATION) * section_header_p->NumberOfRelocations
      ) )
  {
    retval = reinterpret_cast<const IMAGE_RELOCATION*>(
               &m_obj_data[section_header_p->PointerToRelocations]
             );
  }
  
  return retval;
}


//=============================================================================


//! ĔzulăZNVf[^ׂB
bool ObjFile::IsEqualSection(
  int           i_section_index,
  const BYTE*   i_data,
  int           i_size
) const
{
  bool                          retval           = false;
  const IMAGE_SECTION_HEADER*   section_header_p = NULL;
  const BYTE*                   section_data_p   = NULL;
  const IMAGE_RELOCATION*       relocations      = NULL;
  int                           reloc_count      = 0;
  size_t                        cmp_index        = 0;
  size_t                        cmp_size         = 0;
  
  do
  {
    // ZNVwb_ւ̃|C^擾B
    section_header_p = GetSectionHeaderP(i_section_index);
    // ZNVf[^ւ̃|C^擾B
    section_data_p = GetSectionDataP(i_section_index);
    if( section_header_p == NULL || section_data_p == NULL )
    {
      _RPTF1(_CRT_WARN,
             "* ERROR * : ZNVCfbNX %d ͗Lł͂܂B\n",
             i_section_index
            );
      break;
    }
    
    // Ĕzu擾B
    reloc_count = GetRelocationCount(i_section_index);
    relocations = GetRelocations(i_section_index);
    if( reloc_count && relocations == NULL )
    {
      _RPTF0(_CRT_WARN, "Ĕzu񂪉Ă܂B\n");
      break;
    }
    
    // Ĕzu񂪂ꍇ͍ŏ̍ĔzũItZbg܂ŔrB
    // Ȃꍇ͈ i_size rB
    cmp_size = reloc_count ? relocations[0].VirtualAddress : i_size;
    if( memcmp(i_data, section_data_p, cmp_size) == 0 )  // ŏ̃ubN
    {
      retval = true;
      // SẴubNׂB
      for(int i = 1; i <= reloc_count; i++)
      {
        cmp_index = relocations[i - 1].VirtualAddress + sizeof(DWORD);
        if( i == reloc_count )
        {
          cmp_size = section_header_p->SizeOfRawData
                   - (relocations[i - 1].VirtualAddress +  sizeof(DWORD))
                   ;
        }
        else
        {
          cmp_size = relocations[i].VirtualAddress
                   - (relocations[i - 1].VirtualAddress +  sizeof(DWORD))
                   ;
        }
        
        if( cmp_size
        &&  memcmp(i_data + cmp_index, section_data_p + cmp_index, cmp_size) != 0 )
        {
          retval = false;
          break;
        }
      }
    }
  }
  while( 0 );
  
  return retval;
}


//=============================================================================


//! i_image_p  i_symbol_index Ԗڂ̃V{TB
BYTE* ObjFile::FindSymbolData(
  int     i_symbol_index,
  BYTE*   i_image_p,
  size_t  i_image_size
) const
{
  BYTE*                         retval            = NULL;
  const IMAGE_SYMBOL*           symbol_p          = NULL;
  const IMAGE_SECTION_HEADER*   section_header_p  = NULL;
  const BYTE*                   section_p         = NULL;
  
  do
  {
    // V{ւ̃|C^擾B
    symbol_p = GetSymbolP(i_symbol_index);
    if( symbol_p == NULL )
    {
      _RPTF1(_CRT_WARN,
             "V{CfbNX %d ͗Lł͂܂B\n", i_symbol_index
            );
      break;
    }
    
    // ZNVwb_̃|C^擾B
    section_header_p = GetSectionHeaderP(symbol_p->SectionNumber - 1);
    if( section_header_p == NULL || section_header_p->SizeOfRawData == 0 )
    {
      _RPTF0(_CRT_WARN, "V{f[^Ă܂B\n");
      break;
    }

    for(size_t i = 0; i < i_image_size - section_header_p->SizeOfRawData; i++)
    {
      if( IsEqualSection(
            symbol_p->SectionNumber - 1,
            i_image_p + i,
            i_image_size - i
          ) )
      {
        retval = i_image_p + i + symbol_p->Value;
        break;
      }
    }
  }
  while( 0 );
  
  return retval;
}


//=============================================================================


